#! /usr/bin/env python 
#
# mc.py 
# P.Clote
#
# Program performs MonteCarlo simulation (Boltzmannian)
#
# Functions defined in this file:
# def initialEmptySecStr(rna):
# def canAddBasePair(i,j,n,L):
# def isLocallyOptimal(rna,secStr,L):
# def kLocallyOptimal(rna,secStr,L,maxNumBP):
# def listOfNextStatesByAddingBasePair(rna,secStr,L):
# def listOfNextStatesByRemovingBasePair(rna,secStr,L):
# def nextConfiguration(rna,secStr,L):
# def monteCarlo(rna,m):
# def main(rnaStringOrFile):


import os, sys, random, copy, math, string, stats
from header import THRESHOLD, AUenergy, GUenergy, GCenergy
from header import TEMP, NUM, NUMRUNS
from misc import getInputRNA, numBasePairsInOptimal, isOptimal
from misc import basePair, validNucleotideString, energy
from misc import printSecStrListAsString, numParenPairs

DEBUG = 0


def initialEmptySecStr(rna):
  n = len(rna)
  L = []
  for i in range(n): L.append(".")
  return L

def canAddBasePair(i,j,n,L):
  #n is length of RNA sequence with secondary structure secStr
  #L is list of base pairs of secStr obtained by function
  #getBasePairList() or obtained incrementally from an empty list
  #Current function tests if 0<= i < j < n
  #and whether for all (x,y) in L it is NOT the case that
  #either  i<x<j<y  or x<i<y<j.
  #WARNING: This function does NOT check that nucleotides in positions
  #i,j can actually base pair. This is done by basePair(rna[i],rna[j])
  assert (0<=i<j<n)
  for (x,y) in L:
    if x==i and y!=j or x!=i and y==j: return 0
    if (i<x<j<y) or (x<i<y<j): return 0
  return 1

    
def listOfNextStatesByAddingBasePair(rna,secStr,L):
  #L is list of base pairs from secStr, a secondary structure for rna
  #WARNING: The argument passed in secStr is a LIST; however, this
  #the current code is correct as well if secStr is a string of (,),.
  basePairsWhichCanBeAdded = []
  n = len(rna)
  for i in range(n-1-THRESHOLD): #NOTE i <= n-2-THRESHOLD, i.e. leave room for j
    if secStr[i]!=".": continue 
    for j in range(i+THRESHOLD+1,n):
      if secStr[j]!=".": continue 
      if basePair(rna[i],rna[j]) and canAddBasePair(i,j,n,L):
        basePairsWhichCanBeAdded.append( (i,j) ) 
  return basePairsWhichCanBeAdded


def isLocallyOptimal(rna,secStr,L):
  return listOfNextStatesByAddingBasePair(rna,secStr,L)==[]
  
def kLocallyOptimal(rna,secStr,L,maxNumBP):
  if not isLocallyOptimal(rna,secStr,L):
    return -1
  else:
    num = numParenPairs(secStr)
    return maxNumBP-num
    
  

 
def listOfNextStatesByRemovingBasePair(rna,secStr,L):
  return L
  #WARNING: Do not inadvertantly change L in this function



def nextConfiguration(rna,secStr,L):
  L0 = listOfNextStatesByAddingBasePair(rna,secStr,L)
  L1 = listOfNextStatesByRemovingBasePair(rna,secStr,L)
  m0 = len(L0); m1 = len(L1)
  tmp = copy.deepcopy(secStr)
  tmpList = copy.deepcopy(L)
  x  = random.randint(0,m0+m1-1)
  if x<m0: #compute tmp sec str by adding base pair (i,j)
    (i,j) = L0[x]
    tmpList.append( (i,j) )
    tmp[i] = '('; tmp[j] = ')'
  else:    #remove base pair (i,j)
    (i,j) = L1[x-m0]
    tmp[i] = '.'; tmp[j] = '.'
    tmpList.remove( (i,j) )
  if DEBUG: printSecStrListAsString(tmp)
  E0 = energy(rna,L)
  E1 = energy(rna,tmpList)
  if E1<E0:
    flag = 1 #flag is boolean variable, with value 1 if must update
  else:
    z = random.uniform(0,1)
    if DEBUG: print E0,E1,z, math.exp( -(E1-E0)/TEMP )
    if z<math.exp( -(E1-E0)/TEMP ):
      flag = 1 
    else:
      flag = 0 
  if flag:
    secStr = tmp
    L      = tmpList
  return (flag,secStr,L)    


def monteCarlo(rna,m):
  # m is numBasePairsInOptimal(rna)
  #NOTE: Both secStr and L are lists
  n      = len(rna)
  D      = {} #D[k] is num times visited k-locally suboptimal
  for i in range(n/2): D[i]=0
  secStr = initialEmptySecStr(rna)
  L      = [] #initially empty list of base pairs
  num    = 0 #count number of Monte Carlo steps
#  while not isOptimal(secStr,m) and num < NUM:
  while not isOptimal(secStr,m):
    (flag,secStr,L) = nextConfiguration(rna,secStr,L)
    k = kLocallyOptimal(rna,secStr,L,m) 
    if k!=-1: D[k] += 1
    if DEBUG:
      print flag,secStr,L
      print rna
      printSecStrListAsString(secStr)
    num += 1
  if isOptimal(secStr,m): D[0]=1
    #Reason for this line is if terminated while loop when num==NUM
  return num,secStr,L,D


def main(rnaStringOrFile):
  rna = getInputRNA(flag,rnaStringOrFile)
  if not validNucleotideString(rna):
    print "Invalid nucleotide string"
    sys.exit(1)
  m      = numBasePairsInOptimal(rna)
  n      = len(rna)
  firstPassageTimeList = []
  DD = {} #DD is master dictionary, DD[k] is initially list of
          #number of times visited k-locally optimal structure
          #Final value of DD[k] is (mean,stdev,max,min) of
          #number of times visited k-locally optimal structure
  for k in range(n/2): DD[k] = []
  for i in range(NUMRUNS):
    num,secStr,L,D = monteCarlo(rna,m)
    firstPassageTimeList.append(num)
    for k in range(n/2): DD[k].append( D[k] )
    del(D) #free up space taken by D
  if DEBUG: print firstPassageTimeList
  print "Statistics after %d runs" % NUMRUNS
  mean0,stdev0,max0,min0 = stats.getStats(firstPassageTimeList)
  print "Mean:%f\tStDev:%f\tMax:%f\tMin:%f" % (mean0,stdev0,max0,min0)  
  for k in range(n/2):
    mean,stdev,max,min = stats.getStats(DD[k])
    DD[k] = (mean,stdev,max,min)
    if max==0: continue
    text = "%d-locally optimal visits\n\tMean:%f\tStDev:%f\tMax:%f\tMin:%f"
    print text % (k,mean,stdev,max,min)
  return mean0,stdev0,max0,min0,DD



 
if __name__ == '__main__': 
  if len(sys.argv)<3:
    text = """
           Usage: %s -s/-f RNA
           1) -s flag: RNA is nucleotide string (upper or lower case)
           2) -f flag: RNA is FASTA format file of nucleotides 
           """
    print text % sys.argv[0]
    sys.exit(1)
  flag = sys.argv[1]
  rnaStringOrFile = sys.argv[2]
  main(rnaStringOrFile)
  

