#! /usr/bin/env python 
#
# misc.py 
# P.Clote
#
# WARNING: This code calls a.out which is executable of
# program rnaSecStr.c. Make sure that THRESHOLD value in the
# current program matches that in rnaSecStr.c before running.
#
# Functions defined in this file:
# def auPair(x,y):
# def gcPair(x,y):
# def guPair(x,y):
# def watsonCrick(x,y):
# def basePair(x,y):
# def energyOfBasePair(x,y):
# def energy(rna,L):
# def printSecStrListAsString(L):
# def getBasePairList(secStr):
# def validNucleotideString(rna):
# def preValidSecStr(rna,secStr):
# def numParenPairs(secStr):
# def isOptimal(secStr,m):
# def getInputRNA(flag,rnaStringOrFile):


import os, sys, random, copy, math, string, stats
from header import THRESHOLD, AUenergy, GUenergy, GCenergy
from header import TEMP, NUM, NUMRUNS


def auPair(x,y):
  #Assume uppercase characters
  return (x,y) in [('A','U'),('U','A'),('A','T'),('T','A')]

def gcPair(x,y):
  #Assume uppercase characters
  return (x,y) in [('G','C'),('C','G')]

def guPair(x,y):
  #Assume uppercase characters
  return (x,y) in [('G','U'),('U','G'),('G','T'),('T','G')]

def watsonCrick(x,y):
  #Assume uppercase characters
  return auPair(x,y) or gcPair(x,y)

def basePair(x,y):
  #Assume uppercase characters
  return watsonCrick(x,y) or guPair(x,y)

def energyOfBasePair(x,y):
  if auPair(x,y): return AUenergy 
  elif guPair(x,y): return GUenergy 
  elif gcPair(x,y): return GCenergy 
  else: return 0

def energy(rna,L):
  #return energy of base pairs in list L
  #This assumes that L is a list of base paired positions in valid
  #secondary structure for rna
  E = 0
  for (x,y) in L: E += energyOfBasePair(rna[x],rna[y])
  return E

def printSecStrListAsString(L):
  #Since Python strings are immutable (ie not arrays as in C)
  #we prefer to consider secondary structures as lists;
  #i.e. instead of ((..)) we have ['(','(','.','.',')',')']
  #Thus this function will print out the list as a string
  s = string.join(L,"")
  print s


def getBasePairList(secStr):
  #Returns list L of ordered pairs (i,j) where i<j and
  #positions i,j occupied by balancing parentheses
  #For linear time efficiency, use stack
  #Assume that secStr is string consisting of '(',')' and '.'
  #Values 1,-1 returned mean NOT well balanced
  # 1 means too many ( with respect to )
  #-1 means too many ) with respect to (
  #If 1,-1 not returned, then return (possibly empty) list
  #is returned
  S = []  #empty stack
  L = []  #initially empty list of base pairs
  n = len(secStr)
  for j in range(n):
    ch = secStr[j]
    if ch == '(': S.append(j)
    elif ch == ')': 
      if S==[]: return -1
      else: 
        i=S.pop(); L.append( (i,j) )
  if S!=[]: return 1
  return L   

def validNucleotideString(rna):
  rna = rna.upper()
  nucleotides = ['A','C','G','T','U']
  for ch in rna:
    if ch not in nucleotides: return 0
  return 1
 

def preValidSecStr(rna,secStr):
  #tests whether secStr is "almost" valid secondary structure of string rna
  #The function getBasePairList() returns a list if its argument is
  #a string which is well-balanced parenthesis expression with dots.
  #Since getBasePairList() assumes its argument contains only (,) and .
  #the current function checks that all symbols in secStr are
  #indeed (,) and . and that there is an equal number of ( and )
  #
  if len(rna)!=len(secStr): return 0
  numLeftParen  = secStr.count('(')
  numRightParen = secStr.count(')')
  numDots       = secStr.count('.')
  if numLeftParen != numRightParen: return 0
  if numLeftParen + numRightParen + numDots != len(rna): return 0
  return 1


def numParenPairs(secStr):
  #Assume that secStr is well-balanced parenthesis string with dots
  return secStr.count('(')

def isOptimal(secStr,m):
  #Assume that m is Nussinov-Jacobson maximum number of base pairs
  return secStr.count('(') == m

def getInputRNA(flag,rnaStringOrFile):
  if flag=="-s":
    rna = rnaStringOrFile
  elif flag=="-f":
    file = open(rnaStringOrFile,"r")
    line = file.readline()
    while line[0]==">": line = file.readline()
    rna = line+file.read()  #read remainder of file
    rna = rna.replace("\n","")
  return rna.upper()

def numBasePairsInOptimal(rna):
  cmd = "./a.out %s" % rna
  file = os.popen(cmd)
  file.readline()  #remove first line which is sequence
  secStr = file.readline()[:-1]
  return secStr.count('(')



def main(L):
  #Main function used to debug code in this file
  print "hi"
  print isOptimal("((.().))",3)

 
if __name__ == '__main__': 
  L = sys.argv
  main(L)

