#! /usr/bin/env python 
#
# wangLandau.py 
# P.Clote
#
# Functions defined in this file:
# def initialEmptySecStr(rna):
# def canAddBasePair(i,j,n,L):
# def isLocallyOptimal(rna,secStr,L):
# def kLocallyOptimal(rna,secStr,L,maxNumBP):
# def listOfNextStatesByAddingBasePair(rna,secStr,L):
# def listOfNextStatesByRemovingBasePair(rna,secStr,L):
# def nextConfiguration(rna,secStr,L):
# def monteCarlo(rna,m):
# def main(rnaStringOrFile):


import os, sys, random, copy, math, string, stats
from header import THRESHOLD, AUenergy, GUenergy, GCenergy
from header import TEMP, NUM, NUMRUNS
from misc import getInputRNA, numBasePairsInOptimal, isOptimal
from misc import basePair, validNucleotideString, energy
from misc import printSecStrListAsString, numParenPairs


DEBUG      = 1
FINE_DEBUG = 1
global FLATNESS_CONSTANT
global r_x 
TMP_DENSITY_FILENAME = 'tmpDensityOfStates'

'''initialEmptySecStr(rna) returns empty structure for given rna seq'''
def initialEmptySecStr(rna):
  n = len(rna)
  L = []
  for i in range(n): L.append(".")
  return L

'''canAddBasePair(i,j,n,L) determines whether a bp can be added in interval (i,j)'''
def canAddBasePair(i,j,n,L):
  #n is length of RNA sequence with secondary structure secStr
  #L is list of base pairs of secStr obtained by function
  #getBasePairList() or obtained incrementally from an empty list
  #Current function tests if 0<= i < j < n
  #and whether for all (x,y) in L it is NOT the case that
  #either  i<x<j<y  or x<i<y<j.
  #WARNING: This function does NOT check that nucleotides in positions
  #i,j can actually base pair. This is done by basePair(rna[i],rna[j])
  assert (0<=i<j<n)
  for (x,y) in L:
    if x==i and y!=j or x!=i and y==j: return 0
    if (i<x<j<y) or (x<i<y<j): return 0
  return 1

    
'''listOfNextStatesByAddingBasePair(rna,secStr,L) returns list of bp that can be added to secStr without inducing pseudoknot'''
def listOfNextStatesByAddingBasePair(rna,secStr,L):
  #L is list of base pairs from secStr, a secondary structure for rna
  #WARNING: The argument passed in secStr is a LIST; however, this
  #current code is correct as well if secStr is a string of (,),.
  basePairsWhichCanBeAdded = []
  n = len(rna)

  for i in range(n-1-THRESHOLD): #NOTE i <= n-2-THRESHOLD, i.e. leave room for j
    if secStr[i]!=".": continue 
    for j in range(i+THRESHOLD+1,n):
      if secStr[j]!=".": continue 
      if basePair(rna[i],rna[j]) and canAddBasePair(i,j,n,L):
        basePairsWhichCanBeAdded.append( (i,j) ) 
  return basePairsWhichCanBeAdded


def listOfNextStatesByRemovingBasePair(rna,secStr,L):
  return L
  #WARNING: Do not inadvertantly change L in this function



def nextConfiguration(rna,secStr,L,g,h,c):
  global r_x 
  
  #g is density of states, h is histogram, c is modification factor
  N  = len(h)-1
  L0 = listOfNextStatesByAddingBasePair(rna,secStr,L)
  L1 = listOfNextStatesByRemovingBasePair(rna,secStr,L)
  m0 = len(L0); m1 = len(L1)
  tmp = copy.deepcopy(secStr)
  tmpList = copy.deepcopy(L)
  
  #print L0
  #print L1
  
  #x = random.randint(0,m0+m1-1)
  r_x = randDebug(r_x)
  x = r_x % (m0+m1)
  print "r_x:%d, x:%d, m0:%d, m1:%d" % (r_x, x, m0, m1)

  if x<m0: #compute tmp sec str by adding base pair (i,j)
    (i,j) = L0[x]
    tmpList.append( (i,j) )
    tmp[i] = '('; tmp[j] = ')'
    #print "add at %d et %d" % (i, j)
  else:    #remove base pair (i,j)
    (i,j) = L1[x-m0]
    tmp[i] = '.'; tmp[j] = '.'
    tmpList.remove( (i,j) )
    #print "remove at %d et %d" % (i, j)
  if FINE_DEBUG: printSecStrListAsString(tmp)
  E0 = energy(rna,L)
  E1 = energy(rna,tmpList)
  x0 = len(L)              #energy bin for starting structure
  x1 = len(tmpList)        #energy bin for neighbor structure
  if FINE_DEBUG: print E0,E1,x0,x1
  
  #z = random.uniform(0,1) #random real in (0,1)
  z = float(randDebug(r_x)/1023.)
  print "z:%g, gx0:%g, gx0/gx1:%g" % (z, g[x0], float(g[x0])/g[x1])
  #print L, tmpList, L0, L1;

  if g[x1]==0 or z<float(g[x0])/g[x1]: #Wang Landau criterion
    flag   = 1
    secStr = tmp
    L      = tmpList
    h[x1] += 1
    h[N]  += 1
    g[x1] *= c
    #print "***GO: h %d add 1, g %d multipe %g" % (x1, x1, c)
  else:
    flag = 0 
    h[x0] += 1
    h[N]  += 1
    g[x0] *= c
    #print "***NO GO: h %d add 1, g %d multipe %g" % (x0, x0, c) 
  return (flag,secStr,L,g,h)    


def flat(h):
  #h is histogram considered as list with energy bins 0,...,N-1
  #h[N] is number of visits altogether, h[i] number of visits of i-th bin
  N = len(h)-1
  if h[N] == 0:
    return 0
  for i in range(N):
    if h[i] < FLATNESS_CONSTANT *h[N]/N:
      return 0
  return 1

def normalizeDensityOfStates(g):
  sum = 0.0
  for i in range(len(g)):
    sum += g[i]
  for i in range(len(g)):
    g[i] = g[i]/sum
 

def printDensityOfStates(g):
  """prints density of states in nice format"""
  print "Num base pairs\tdensity"
  for i in range(len(g)):
    print "%d\t\t%g" % (i,g[i])

def wangLandau(rna,m):
  # m is numBasePairsInOptimal(rna)
  #NOTE: Both secStr and L are lists
  n      = len(rna)
  N      = m+1                 # energy bins 0,1,...,m 
  num    = 0                   #count number of Monte Carlo steps
  g      = []
  
  #print N, N+1
  
  for i in range(N): 
    g.append(1.0/N)            #initialize density of states   
  c = math.e                   #c is modification factor
  epsilon = 0.001              #terminate when c<=(1+epsilon); about 10 iterations for 0.001
  M       = 100                #number of iter between checking if h is flat
  secStr  = initialEmptySecStr(rna)  #initial sec str of rna
  L       = []                 #initially empty list of base pairs
  tmperr  = open(TMP_DENSITY_FILENAME,'w') #output for tmp density of states
  while c>(1+epsilon):
    h  = []
    for i in range(N): 
      h.append(0)             #reset histogram to empty
    h.append(0)               #empty histogram; h[N]=0
    print secStr, L, g, h
    
    while not flat(h):
      for iterationCount in range(M):
        (flag,secStr,L,g,h) = nextConfiguration(rna,secStr,L,g,h,c)
        raw_input("next: ")
      normalizeDensityOfStates(g)   #normalize density of states after M moves
      num   += M            #increment num Monte Carlo steps
      if DEBUG:
        print flag,secStr,L
        print rna
        printSecStrListAsString(secStr)
        print "Density of states: %s" % str(g)
        print "Histogram: %s" % str(h)
        print "Flat?: %d" % flat(h)
      if FINE_DEBUG and num%1000==0:
        tmperr.write("\nModification factor: %f\n" % c)
        tmperr.write("Number of MonteCarlo moves: %d\n" % num)
        tmperr.write("Num base pairs\tdensity\n")
        tmperr.write("Histogram: %s\n" % str(h))
        for i in range(len(g)):
          tmperr.write("%d\t\t%f\n" % (i,g[i]))
    
    c = c**(0.5)
    if DEBUG: print "Modification factor changed to: %f" % c
    
  tmperr.close()
  return num,g,h

def randDebug(seed):
  return (13*seed+1) % 1023

def main(rnaStringOrFile):
  global r_x 
  
  rna = getInputRNA(flag,rnaStringOrFile)
  if not validNucleotideString(rna):
    print "Invalid nucleotide string"
    sys.exit(1)
  m      = numBasePairsInOptimal(rna)
  n      = len(rna)
  r_x      = 1
  
  num,g,h = wangLandau(rna,m)
  print "Number of MonteCarlo moves: %d" % num
  printDensityOfStates(g)

 
if __name__ == '__main__': 
  if len(sys.argv)<3:
    text = """
           Usage: %s -s/-f RNA [flatnessConstant]
           1) -s flag: RNA is nucleotide string (upper or lower case)
           2) -f flag: RNA is FASTA format file of nucleotides 
           3) flatness constant in (0,1), default is 0.8
           """
    print text % sys.argv[0]
    sys.exit(1)
  flag = sys.argv[1]
  rnaStringOrFile = sys.argv[2]
  if len(sys.argv)>3: #WARNING - FLATNESS_CONSTANT is global variable
    FLATNESS_CONSTANT = float(sys.argv[3])
  else:
    FLATNESS_CONSTANT = 0.8
  main(rnaStringOrFile)
  

